import numpy as np
import pandas as pd
from pyDOE import lhs
from pyrvea.Problem.baseproblem import BaseProblem


class OptTestFunctions:

    """Test functions for single/multi-objective problems to test
    the performance of evolutionary algorithms.

    See: https://en.wikipedia.org/wiki/Test_functions_for_optimization

    Parameters
    ----------
    name : str
        name of the test function
    num_of_variables : int
        number of decision variables
    num_of_objectives : int
        number of objectives
    num_of_constraints : int
        number of constraints
    upper_limits : float
        upper boundaries for test data
    lower_limits : float
        lower boundaries for test data
    """

    def __init__(self, name=None, num_of_variables=None):
        self.name = name
        self.num_of_variables = num_of_variables
        # Define search domain for test functions
        test_f_params = {
            # Single objective functions
            "Sphere": {"var": 2, "obj": 1, "bounds": (-5, 5)},
            "Matyas": {"var": 2, "obj": 1, "bounds": (-10, 10)},
            "Himmelblau": {"var": 2, "obj": 1, "bounds": (-5, 5)},
            "Rastigrin": {"var": 2, "obj": 1, "bounds": (-5.12, 5.12)},
            "Three-hump camel": {"var": 2, "obj": 1, "bounds": (-5, 5)},
            "Goldstein-Price": {"var": 2, "obj": 1, "bounds": (-2, 2)},
            "LeviN13": {"var": 2, "obj": 1, "bounds": (-10, 10)},
            "SchafferN2": {"var": 2, "obj": 1, "bounds": (-100, 100)},
            # Multi-objective functions
            "Coello_ex1": {"var": 2, "obj": 2, "bounds": (0, 1)},
            "Fonseca-Fleming": {"var": 2, "obj": 2, "bounds": (-4, 4)},
            "Kursawe": {"var": 2, "obj": 2, "bounds": (-5, 5)},
            "SchafferN1": {"var": 1, "obj": 2, "bounds": (-100, 100)},
        }

        if self.name in test_f_params.keys():
            if self.num_of_variables is None:
                self.num_of_variables = test_f_params[self.name]["var"]
            self.num_of_objectives = test_f_params[self.name]["obj"]
            self.lower_limits = test_f_params[self.name]["bounds"][0]
            self.upper_limits = test_f_params[self.name]["bounds"][1]

    def __call__(self, x):
        return self.objectives(x)

    def objectives(self, decision_variables) -> list:
        """Use this method to calculate objective functions.

        Parameters
        ----------
        decision_variables : np.ndarray
            The decision variables.

        Returns
        -------
        The objective functions.

        """
        # Single objective test functions

        if self.name == "Sphere":
            # Sphere function, -5 <= x <= 5

            x = np.asarray(decision_variables)
            self.obj_func = sum(x ** 2)

        elif self.name == "Matyas":
            # Matyas function, -10 <= x, y <= 10

            x = np.asarray(decision_variables[0])
            y = np.asarray(decision_variables[1])
            self.obj_func = 0.26 * (x ** 2 + y ** 2) - 0.48 * x * y

        elif self.name == "Himmelblau":
            # Himmelblau's function, -5 <= x, y <= 5

            x = np.asarray(decision_variables[0])
            y = np.asarray(decision_variables[1])
            self.obj_func = (x ** 2 + y - 11) ** 2 + (x + y ** 2 - 7) ** 2

        elif self.name == "Rastigrin":
            # Rastigrin function, -5.12 <= x <= 5.12

            x = np.asarray(decision_variables)
            n = len(x)
            self.obj_func = 10 * n + sum(x ** 2 - 10 * np.cos(2 * np.pi * x))

        elif self.name == "Three-hump camel":
            # Three-hump camel function,  -5 <= x, y <= 5

            x = np.asarray(decision_variables[0])
            y = np.asarray(decision_variables[1])
            self.obj_func = 2 * x ** 2 - 1.05 * x ** 4 + (x ** 6) / 6 + x * y + y ** 2

        elif self.name == "Goldstein-Price":
            # Goldstein-Price function, -2 <= x, y <= 2

            x = np.asarray(decision_variables[0])
            y = np.asarray(decision_variables[1])
            self.obj_func = (
                1
                + (x + y + 1) ** 2
                * (19 - 14 * x + 3 * (x ** 2) - 14 * y + 6 * x * y + 3 * (y ** 2))
            ) * (
                30
                + (2 * x - 3 * y) ** 2
                * (18 - 32 * x + 12 * (x ** 2) + 48 * y - 36 * x * y + 27 * (y ** 2))
            )

        elif self.name == "LeviN13":
            # Levi function N.13, -10 <= x, y <= 10

            x = np.asarray(decision_variables[0])
            y = np.asarray(decision_variables[1])
            self.obj_func = (
                np.sin(3 * np.pi * x) ** 2
                + (x - 1) ** 2 * (1 + np.sin(3 * np.pi * y) ** 2)
                + (y - 1) ** 2 * (1 + np.sin(2 * np.pi * y) ** 2)
            )

        elif self.name == "SchafferN2":
            # Schaffer function N. 2, -100 <= x, y <= 100

            x = np.asarray(decision_variables[0])
            y = np.asarray(decision_variables[1])
            self.obj_func = (
                0.5
                + (np.sin((x ** 2 - y ** 2) ** 2) - 0.5)
                / (1 + 0.001 * (x ** 2 + y ** 2)) ** 2
            )

        # Test functions for multi-objective optimization

        elif self.name == "Coello_ex1":
            x = np.asarray(decision_variables[0])
            y = np.asarray(decision_variables[1])
            a = 2
            q = 4
            f1 = x
            f2 = (1 + 10 * y) * (
                1
                - (x / (1 + 10 * y)) ** a
                - x / (1 + 10 * y) * np.sin(2 * np.pi * q * x)
            )

            self.obj_func = [f1, f2]

        elif self.name == "Kursawe":
            x1 = np.asarray(decision_variables[0])
            x2 = np.asarray(decision_variables[1])
            x3 = np.asarray(decision_variables[1])
            f1 = -10 * np.exp(-0.2 * np.sqrt(x1 ** 2 + x2 ** 2)) - 10 * np.exp(
                -0.2 * np.sqrt(x2 ** 2 + x3 ** 2)
            )
            f2 = (
                abs(x1) ** 0.8
                + 5.0 * np.sin(x1 ** 3)
                + abs(x2) ** 0.8
                + 5.0 * np.sin(x2 ** 3)
                + abs(x3) ** 0.8
                + 5.0 * np.sin(x3 ** 3)
            )
            self.obj_func = [f1, f2]

        elif self.name == "Fonseca-Fleming":
            f1 = 0
            f2 = 0

            for n in range(len(decision_variables)):
                f1 += (
                    decision_variables[n] - 1 / np.sqrt(len(decision_variables))
                ) ** 2
                f2 += (
                    decision_variables[n] + 1 / np.sqrt(len(decision_variables))
                ) ** 2

            f1 = 1 - np.exp(-f1)
            f2 = 1 - np.exp(-f2)

            self.obj_func = [f1, f2]

        elif self.name == "SchafferN1":
            x = np.asarray(decision_variables[0])

            f1 = x ** 2
            f2 = (x - 2) ** 2

            self.obj_func = [f1, f2]

        return self.obj_func

    def create_training_data(self, samples=150, method="random", seed=None):
        """Create training data for test functions.

        Parameters
        ----------
        samples : int
            number of samples
        method : str
            method to use in data creation. Possible values random, lhs, linear,
            linear+zeros, linear+reverse.
        seed : int
            if a number is given, random data will be seeded
        """

        np.random.seed(seed)
        training_data_input = None

        if method == "random":

            training_data_input = np.random.uniform(
                self.lower_limits, self.upper_limits, (samples, self.num_of_variables)
            )

        elif method == "lhs":

            training_data_input = lhs(self.num_of_variables, samples) * (
                abs(self.upper_limits) + abs(self.lower_limits)
            ) - abs(self.upper_limits)

        elif method == "linear":

            training_data_input = np.linspace(
                (self.lower_limits, self.lower_limits),
                (self.upper_limits, self.upper_limits),
                samples,
            )

        elif method == "linear+zeros":

            training_data_input = np.linspace(
                (self.lower_limits, self.lower_limits),
                (self.upper_limits, self.upper_limits),
                samples,
            )
            tmp = np.linspace(self.upper_limits, 0, samples)
            x2 = np.zeros_like(tmp)
            training_data_input = np.vstack(
                (np.hstack((tmp, tmp)), np.hstack((tmp, x2)))
            ).T

        elif method == "linear+reverse":

            np.linspace(
                (self.lower_limits, self.lower_limits),
                (self.upper_limits, self.upper_limits),
                samples,
            )
            tmp = np.linspace(self.upper_limits, 0, samples)
            x2 = np.flip(tmp)
            training_data_input = np.vstack(
                (np.hstack((tmp, tmp)), np.hstack((tmp, x2)))
            ).T

        training_data_output = np.asarray(
            [self.objectives(x) for x in training_data_input]
        )

        # Convert numpy array into pandas dataframe, and make columns for it
        data = np.hstack((training_data_input, training_data_output))
        dataset = pd.DataFrame.from_records(data)
        x = []
        y = []
        for var in range(training_data_input.shape[1]):
            x.append("x" + str(var + 1))
        for obj in range(training_data_output.shape[1]):
            y.append("f" + str(obj + 1))
        dataset.columns = x + y

        np.random.seed(None)

        return dataset, x, y



from typing import Dict, Union

from desdeo_emo.EAs.BaseEA import BaseDecompositionEA, eaError
from desdeo_emo.population.Population import Population

# from desdeo_emo.selection.APD_Select import APD_Select
from desdeo_emo.selection.APD_Select_constraints import APD_Select
from desdeo_emo.selection.oAPD import Optimistic_APD_Select
from desdeo_emo.selection.robust_APD import robust_APD_Select
from desdeo_problem.Problem import MOProblem


class RVEA(BaseDecompositionEA):
    """The python version reference vector guided evolutionary algorithm.

    Most of the relevant code is contained in the super class. This class just assigns
    the APD selection operator to BaseDecompositionEA.

    NOTE: The APD function had to be slightly modified to accomodate for the fact that
    this version of the algorithm is interactive, and does not have a set termination
    criteria. There is a time component in the APD penalty function formula of the type:
    (t/t_max)^alpha. As there is no set t_max, the formula has been changed. See below,
    the documentation for the argument: penalty_time_component

    See the details of RVEA in the following paper

    R. Cheng, Y. Jin, M. Olhofer and B. Sendhoff, A Reference Vector Guided
    Evolutionary Algorithm for Many-objective Optimization, IEEE Transactions on
    Evolutionary Computation, 2016

    Parameters
    ----------
    problem : MOProblem
        The problem class object specifying the details of the problem.
    population_size : int, optional
        The desired population size, by default None, which sets up a default value
        of population size depending upon the dimensionaly of the problem.
    population_params : Dict, optional
        The parameters for the population class, by default None. See
        desdeo_emo.population.Population for more details.
    initial_population : Population, optional
        An initial population class, by default None. Use this if you want to set up
        a specific starting population, such as when the output of one EA is to be
        used as the input of another.
    alpha : float, optional
        The alpha parameter in the APD selection mechanism. Read paper for details.
    lattice_resolution : int, optional
        The number of divisions along individual axes in the objective space to be
        used while creating the reference vector lattice by the simplex lattice
        design. By default None
    a_priori : bool, optional
        A bool variable defining whether a priori preference is to be used or not.
        By default False
    interact : bool, optional
        A bool variable defining whether interactive preference is to be used or
        not. By default False
    n_iterations : int, optional
        The total number of iterations to be run, by default 10. This is not a hard
        limit and is only used for an internal counter.
    n_gen_per_iter : int, optional
        The total number of generations in an iteration to be run, by default 100.
        This is not a hard limit and is only used for an internal counter.
    total_function_evaluations :int, optional
        Set an upper limit to the total number of function evaluations. When set to
        zero, this argument is ignored and other termination criteria are used.
    penalty_time_component: Union[str, float], optional
        The APD formula had to be slightly changed.
        If penalty_time_component is a float between [0, 1], (t/t_max) is replaced by
        that constant for the entire algorithm.
        If penalty_time_component is "original", the original intent of the paper is
        followed and (t/t_max) is calculated as
        (current generation count/total number of generations).
        If penalty_time_component is "function_count", (t/t_max) is calculated as
        (current function evaluation count/total number of function evaluations)
        If penalty_time_component is "interactive", (t/t_max)  is calculated as
        (Current gen count within an iteration/Total gen count within an iteration).
        Hence, time penalty is always zero at the beginning of each iteration, and one
        at the end of each iteration.
        Note: If the penalty_time_component ever exceeds one, the value one is used as
        the penalty_time_component.
        If no value is provided, an appropriate default is selected.
        If `interact` is true, penalty_time_component is "interactive" by default.
        If `interact` is false, but `total_function_evaluations` is provided,
        penalty_time_component is "function_count" by default.
        If `interact` is false, but `total_function_evaluations` is not provided,
        penalty_time_component is "original" by default.
    """

    def __init__(
        self,
        problem: MOProblem,
        population_size: int = None,
        population_params: Dict = None,
        initial_population: Population = None,
        alpha: float = 2,
        lattice_resolution: int = None,
        selection_type: str = None,
        a_priori: bool = False,
        interact: bool = False,
        use_surrogates: bool = False,
        n_iterations: int = 10,
        n_gen_per_iter: int = 100,
        total_function_evaluations: int = 0,
        time_penalty_component: Union[str, float] = None,
    ):
        super().__init__(
            problem=problem,
            population_size=population_size,
            population_params=population_params,
            initial_population=initial_population,
            lattice_resolution=lattice_resolution,
            a_priori=a_priori,
            interact=interact,
            use_surrogates=use_surrogates,
            n_iterations=n_iterations,
            n_gen_per_iter=n_gen_per_iter,
            total_function_evaluations=total_function_evaluations,
        )
        self.time_penalty_component = time_penalty_component
        time_penalty_component_options = ["original", "function_count", "interactive"]
        if time_penalty_component is None:
            if interact is True:
                time_penalty_component = "interactive"
            elif total_function_evaluations > 0:
                time_penalty_component = "function_count"
            else:
                time_penalty_component = "original"
        if not (type(time_penalty_component) is float or str):
            msg = (
                f"type(time_penalty_component) should be float or str"
                f"Provided type: {type(time_penalty_component)}"
            )
            eaError(msg)
        if type(time_penalty_component) is float:
            if (time_penalty_component <= 0) or (time_penalty_component >= 1):
                msg = (
                    f"time_penalty_component should either be a float in the range"
                    f"[0, 1], or one of {time_penalty_component_options}.\n"
                    f"Provided value = {time_penalty_component}"
                )
                eaError(msg)
            time_penalty_function = self._time_penalty_constant
        if type(time_penalty_component) is str:
            if time_penalty_component == "original":
                time_penalty_function = self._time_penalty_original
            elif time_penalty_component == "function_count":
                time_penalty_function = self._time_penalty_function_count
            elif time_penalty_component == "interactive":
                time_penalty_function = self._time_penalty_interactive
            else:
                msg = (
                    f"time_penalty_component should either be a float in the range"
                    f"[0, 1], or one of {time_penalty_component_options}.\n"
                    f"Provided value = {time_penalty_component}"
                )
                eaError(msg)
        self.time_penalty_function = time_penalty_function
        self.alpha = alpha
        self.selection_type = selection_type
        selection_operator = APD_Select(
            pop=self.population,
            time_penalty_function=self.time_penalty_function,
            alpha=alpha,
            selection_type=selection_type,
        )
        self.selection_operator = selection_operator

    def _time_penalty_constant(self):
        """Returns the constant time penalty value.
        """
        return self.time_penalty_component

    def _time_penalty_original(self):
        """Calculates the appropriate time penalty value, by the original formula.
        """
        return self._current_gen_count / self.total_gen_count

    def _time_penalty_interactive(self):
        """Calculates the appropriate time penalty value.
        """
        return self._gen_count_in_curr_iteration / self.n_gen_per_iter

    def _time_penalty_function_count(self):
        """Calculates the appropriate time penalty value.
        """
        return self._function_evaluation_count / self.total_function_evaluations


class oRVEA(RVEA):
    def __init__(
        self,
        problem: MOProblem,
        population_size: int = None,
        population_params: Dict = None,
        initial_population: Population = None,
        alpha: float = 2,
        lattice_resolution: int = None,
        a_priori: bool = False,
        interact: bool = False,
        use_surrogates: bool = False,
        n_iterations: int = 10,
        n_gen_per_iter: int = 100,
        total_function_evaluations: int = 0,
        time_penalty_component: Union[str, float] = None,
    ):
        super().__init__(
            problem=problem,
            population_size=population_size,
            population_params=population_params,
            initial_population=initial_population,
            lattice_resolution=lattice_resolution,
            a_priori=a_priori,
            interact=interact,
            use_surrogates=use_surrogates,
            n_iterations=n_iterations,
            n_gen_per_iter=n_gen_per_iter,
            total_function_evaluations=total_function_evaluations,
        )
        selection_operator = Optimistic_APD_Select(
            self.population, self.time_penalty_function, alpha
        )
        self.selection_operator = selection_operator


class robust_RVEA(RVEA):
    def __init__(
        self,
        problem: MOProblem,
        population_size: int = None,
        population_params: Dict = None,
        initial_population: Population = None,
        alpha: float = 2,
        lattice_resolution: int = None,
        a_priori: bool = False,
        interact: bool = False,
        use_surrogates: bool = False,
        n_iterations: int = 10,
        n_gen_per_iter: int = 100,
        total_function_evaluations: int = 0,
        time_penalty_component: Union[str, float] = None,
    ):
        super().__init__(
            problem=problem,
            population_size=population_size,
            population_params=population_params,
            initial_population=initial_population,
            lattice_resolution=lattice_resolution,
            a_priori=a_priori,
            interact=interact,
            use_surrogates=use_surrogates,
            n_iterations=n_iterations,
            n_gen_per_iter=n_gen_per_iter,
            total_function_evaluations=total_function_evaluations,
        )
        selection_operator = robust_APD_Select(
            self.population, self.time_penalty_function, alpha
        )
        self.selection_operator = selection_operator



import numpy as np
import plotly
import plotly.graph_objs as go
from scipy.optimize import lsq_linear
from scipy.special import expit

from pyrvea.EAs.PPGA import PPGA
from pyrvea.Population.Population import Population
from pyrvea.Problem.baseproblem import BaseProblem


class EvoNN(BaseProblem):
    """Creates Artificial Neural Network (ANN) models for the EvoNN algorithm.

    These models contain only one hidden node layer. The lower part of the network
    is optimized by a genetic algorithm, and the upper part is optimized by Linear
    Least Square
    algorithm by default.

    Parameters
    ----------
    name : str
        Name of the problem.
    X_train : np.ndarray
        Training data input.
    y_train : np.ndarray
        Training data target values.
    num_of_objectives : int
        The number of objectives.
    params : dict
        Parameters for model training.
    num_samples : int
        The number of data points, or samples.

    Notes
    -----
    The algorithm has been created earlier in MATLAB, and this Python implementation
    has been using
    that code as a basis.

    Python code has been written by Niko Rissanen under the supervision of professor
    Nirupam Chakraborti.

    References
    ----------
    [1] Chakraborti, N. (2014). Strategies for evolutionary data driven modeling in
    chemical and metallurgical Systems.
    In Applications of Metaheuristics in Process Engineering (pp. 89-122). Springer,
    Cham.
    [2] Pettersson, F., Chakraborti, N., & Saxén, H. (2007). A genetic algorithms based
    multi-objective neural net
    applied to noisy blast furnace data. Applied Soft Computing, 7(1), 387-397.

    """

    def __init__(
        self,
        name=None,
        X_train=None,
        y_train=None,
        num_of_objectives=2,
        params=None,
        num_samples=None,
    ):
        super().__init__()

        self.name = name
        self.X_train = X_train
        self.y_train = y_train
        self.num_of_objectives = num_of_objectives
        self.params = params
        self.num_samples = num_samples

    def objectives(self, decision_variables) -> list:
        """ Use this method to calculate objective functions.

        Parameters
        ----------
        decision_variables : np.ndarray
            Variables from the neural network.

        Returns
        -------
        obj_func : list
            The objective function.

        """

        activated_layer, complexity = self.activation(decision_variables)
        _, _, training_error = self.calculate_linear(activated_layer)

        obj_func = [training_error, complexity]

        return obj_func

    def activation(self, non_linear_layer):
        """ Calculates the dot product and applies the activation function.
        Also get complexity for the lower part of the network.

        Parameters
        ----------
        non_linear_layer : np.ndarray
            Weight matrix of the neural network.

        Returns
        -------
        activated_layer : np.ndarray
            The activated non-linear layer before the output.
        complexity : int
            The model's complexity.

        """
        # Calculate the dot product + bias
        out = np.dot(self.X_train, non_linear_layer[1:, :]) + non_linear_layer[0]
        complexity = np.count_nonzero(non_linear_layer)
        activated_layer = self.activate(self.params["activation_func"], out)

        return activated_layer, complexity

    def calculate_linear(self, non_linear_layer):
        """ Apply the linear function to the activated layer
        and calculate the training error.

        Parameters
        ----------
        non_linear_layer : np.ndarray
            Output of the activation function

        Returns
        -------
        linear_layer : np.ndarray
            The optimized weight matrix of the upper part of the network
        predicted_values : np.ndarray
            The prediction of the model
        training_error : float
            The model's training error
        """

        linear_layer = None
        training_error = None

        if self.params["opt_func"] == "llsq":
            linear_solution = np.linalg.lstsq(
                non_linear_layer, self.y_train, rcond=None
            )
            linear_layer = linear_solution[0]

        elif self.params["opt_func"] == "llsq_constrained":
            linear_layer = lsq_linear(
                non_linear_layer, self.y_train, method="bvls", bounds=(0, 1)
            ).x

        predicted_values = np.dot(non_linear_layer, linear_layer)

        if self.params["loss_func"] == "root_mean_square":
            training_error = np.sqrt(np.mean(((self.y_train - predicted_values) ** 2)))

        elif self.params["loss_func"] == "root_median_square":
            training_error = np.sqrt(
                np.median(((self.y_train - predicted_values) ** 2))
            )

        return linear_layer, predicted_values, training_error

    def information_criterion(self, decision_variables):
        """Calculate the information criterion.

        Currently supports Akaike and corrected Akaike Information Criterion.

        Returns
        -------
        Corrected Akaike Information criterion
        """
        activated_layer, complexity = self.activation(decision_variables)
        linear_layer, predicted_values, _ = self.calculate_linear(activated_layer)
        rss = np.sum(((predicted_values - self.y_train) ** 2))
        k = complexity + np.count_nonzero(linear_layer)
        aic = 2 * k + self.num_samples * np.log(rss / self.num_samples)
        aicc = aic + (2 * k * (k + 1) / (self.num_samples - k - 1))

        return aicc

    def select(self, pop, non_dom_front, selection="min_error"):
        """ Select target model from the population.

        Parameters
        ----------
        pop : obj
            The population object.
        non_dom_front : list
            Indices of the models on the non-dominated front.
        selection : str
            The criterion to use for selecting the model.
            Possible values: 'min_error', 'akaike_corrected', 'manual'

        Returns
        -------
        The selected model
        """
        model = None
        fitness = None
        if selection == "min_error":
            # Return the model with the lowest error

            lowest_error = np.argmin(pop.objectives[:, 0])
            model = pop.individuals[lowest_error]
            fitness = pop.fitness[lowest_error]

        elif selection == "akaike_corrected":

            # Calculate Akaike information criterion for the non-dominated front
            # and return the model with the lowest value

            info_c_rank = []

            for i in non_dom_front:

                info_c = self.information_criterion(pop.individuals[i])
                info_c_rank.append((info_c, i))

            info_c_rank.sort()

            model = pop.individuals[info_c_rank[0][1]]
            fitness = pop.fitness[info_c_rank[0][1]]

        elif selection == "manual":

            pareto = pop.objectives[non_dom_front]
            hover = pop.objectives.tolist()

            for i, x in enumerate(hover):
                x.insert(0, "Model " + str(i) + "<br>")

            trace0 = go.Scatter(
                x=pop.objectives[:, 0], y=pop.objectives[:, 1], mode="markers"
            )
            trace1 = go.Scatter(
                x=pareto[:, 0],
                y=pareto[:, 1],
                text=hover,
                hoverinfo="text",
                mode="markers+lines",
            )
            data = [trace0, trace1]
            layout = go.Layout(
                xaxis=dict(title="training error"), yaxis=dict(title="complexity")
            )
            plotly.offline.plot(
                {"data": data, "layout": layout},
                filename=self.name + "_training_models_" + "pareto" + ".html",
                auto_open=True,
            )

            model_idx = None
            while model_idx not in pop.objectives:
                usr_input = input(
                    "Please input the number of the model of your preference: "
                )
                try:
                    model_idx = int(usr_input)
                except ValueError:
                    print("Invalid input, please enter the model number.")
                    continue

                if model_idx not in pop.objectives:
                    print("Model " + str(model_idx) + " not found.")

            model = pop.individuals[int(model_idx)]
            fitness = pop.fitness[int(model_idx)]

        return model, fitness

    @staticmethod
    def activate(name, x):
        if name == "sigmoid":
            return expit(x)

        if name == "relu":
            return np.maximum(x, 0)

        if name == "tanh":
            return np.tanh(x)


class EvoNNModel(EvoNN):
    """The class for the EvoNN surrogate model.

    Parameters
    ----------
    model_parameters : dict
        Parameters passed for the model.
    ea_parameters : dict
        Parameters passed for the genetic algorithm.

    Attributes
    ----------
    name : str
        Name of the model.
    non_linear_layer : np.ndarray
        The weight matrix of the lower part of the network.
    linear_layer : np.ndarray
        The linear layer of the upper part of the network.
    fitness : list
        Fitness of the trained model.
    svr : np.ndarray
        Single variable response of the model.
    log : file
        If logging set to True in params, external log file is stored here.

    """

    def __init__(self, model_parameters=None, ea_parameters=None):
        super().__init__()
        self.name = "EvoNN_Model"
        self.non_linear_layer = None
        self.linear_layer = None
        self.fitness = None
        self.svr = None
        self.log = None
        self.ea_params = ea_parameters
        if model_parameters:
            self.set_params(**model_parameters)
        else:
            self.set_params()

    def set_params(
        self,
        name="EvoNN_Model",
        training_algorithm=PPGA,  # done
        pop_size=500,  # done
        num_nodes=20,  # done
        prob_omit=0.2,  # done
        w_low=-5.0,  # done
        w_high=5.0,  # done
        activation_func="sigmoid",  # done
        opt_func="llsq",  # done
        loss_func="root_median_square",  # done
        selection="akaike_corrected",  # done
        recombination_type="evonn_xover_mutation",  # done
        crossover_type="standard",  # done
        mutation_type="gaussian",  # done
        logging=False,
        plotting=False,
    ):

        """ Set parameters for the EvoNN model.

        Parameters
        ----------
        name : str
            Name of the problem.
        training_algorithm : EA
            Which evolutionary algorithm to use for training the models.
        pop_size : int
            Population size.
        num_nodes : int
            Maximum number of nodes per layer.
        prob_omit : float
            Probability of setting some weights to zero initially.
        w_low : float
            The lower bound for randomly generated weights.
        w_high : float
            The upper bound for randomly generated weights.
        activation_func : str
            Function to use for activation.
        opt_func : str
            Function to use for optimizing the final layer of the model.
        loss_func : str
            The loss function to use.
        selection : str
            The selection to use for selecting the model.
        recombination_type, crossover_type, mutation_type : str or None
            Recombination functions. If recombination_type is specified, crossover and
            mutation
            will be handled by the same function. If None, they are done separately.
        logging : bool
            True to create a logfile, False otherwise.
        plotting : bool
            True to create a plot, False otherwise.
        """

        params = {
            "name": name,
            "training_algorithm": training_algorithm,
            "pop_size": pop_size,
            "num_nodes": num_nodes,
            "prob_omit": prob_omit,
            "w_low": w_low,
            "w_high": w_high,
            "activation_func": activation_func,
            "opt_func": opt_func,
            "loss_func": loss_func,
            "selection": selection,
            "recombination_type": recombination_type,
            "crossover_type": crossover_type,
            "mutation_type": mutation_type,
            "logging": logging,
            "plotting": plotting,
        }

        self.name = name
        self.params = params

    def fit(self, training_data, target_values):
        """Fit data in EvoNN model.

        Parameters
        ----------
        training_data : pd.DataFrame, shape = (numbers of samples, number of variables)
            Training data.
        target_values : pd.DataFrame
            Target values.

        Returns
        -------
        self : returns an instance of self.

        """

        self.X_train = np.asarray(training_data)
        self.y_train = np.asarray(target_values)
        self.num_samples = target_values.shape[0]
        self.num_of_variables = training_data.shape[1]

        self.train()

        self.single_variable_response(ploton=self.params["plotting"])

        if self.params["logging"]:
            self.create_logfile()

        return self

    def train(self):
        """Trains the networks and selects the best model from the non dominated front.

        """
        pop = Population(
            self,
            assign_type="EvoNN",
            pop_size=self.params["pop_size"],
            plotting=self.params["plotting"],
            recombination_type=self.params["recombination_type"],
            crossover_type=self.params["crossover_type"],
            mutation_type=self.params["mutation_type"],
        )
        pop.evolve(EA=self.params["training_algorithm"], ea_parameters=self.ea_params)

        non_dom_front = pop.non_dominated()
        self.non_linear_layer, self.fitness = self.select(
            pop, non_dom_front, self.params["selection"]
        )

        activated_layer, _ = self.activation(self.non_linear_layer)
        self.linear_layer, *_ = self.calculate_linear(activated_layer)

    def predict(self, decision_variables):
        """Predict using the EvoNN model.

        Parameters
        ----------
        decision_variables : pd.DataFrame
            The decision variables used for prediction.

        Returns
        -------
        y : np.ndarray
            The prediction of the model.

        """
        out = (
            np.dot(np.asarray(decision_variables), self.non_linear_layer[1:, :])
            + self.non_linear_layer[0]
        )

        non_linear_layer = self.activate(self.params["activation_func"], out)  # rename

        y = np.dot(non_linear_layer, self.linear_layer)

        return y

    def plot(self, prediction, target, name=None):
        """Creates and shows a plot for the model's prediction.

        Parameters
        ----------
        prediction : np.ndarray
            The prediction of the model.
        target : pd.DataFrame
            The target values.
        name : str
            Filename to save the plot as.
        """
        target = np.asarray(target)
        if name is None:
            name = self.name

        trace0 = go.Scatter(x=prediction, y=target, name="prediction", mode="markers")
        trace1 = go.Scatter(x=target, y=target, name="target")
        data = [trace0, trace1]
        plotly.offline.plot(
            data,
            filename="Tests/"
            + self.params["training_algorithm"].__name__
            + self.__class__.__name__
            + name
            + "_var"
            + str(self.num_of_variables)
            + "_nodes"
            + str(self.params["num_nodes"])
            + ".html",
            auto_open=True,
        )

    def create_logfile(self, name=None):
        """Create a log file containing the parameters for training the model and the EA.

        Parameters
        ----------
        name : str
            Filename to save the log as.

        Returns
        -------
        log_file : file
            An external log file.
        """

        if name is None:
            name = self.name

        # Save params to log file
        log_file = open(
            "Tests/"
            + self.params["training_algorithm"].__name__
            + self.__class__.__name__
            + name
            + "_var"
            + str(self.num_of_variables)
            + "_nodes"
            + str(self.params["num_nodes"])
            + ".log",
            "a",
        )

        for i in self.params:
            print("", i, ":", self.params[i], file=log_file)

        if self.fitness is not None:
            print("fitness: " + str(self.fitness), file=log_file)

        if self.svr is not None:
            print("single variable response: " + str(self.svr), file=log_file)

        return log_file

    def single_variable_response(self, ploton=False):
        """Get the model's response to a single variable.

        Parameters
        ----------
        ploton : bool
            Create and show plot on/off.
        """

        trend = np.loadtxt("trend")
        avg = np.ones((1, self.num_of_variables)) * (np.finfo(float).eps + 1) / 2
        svr = np.empty((0, 2))
        variables = np.ones((len(trend), 1)) * avg

        for i in range(self.num_of_variables):

            variables[:, i] = trend

            out = self.predict(variables)

            if min(out) == max(out):
                out = 0.5 * np.ones(out.size)
            else:
                out = (out - min(out)) / (max(out) - min(out))

            if ploton:
                trace0 = go.Scatter(
                    x=np.arange(len(variables[:, 1])), y=variables[:, i], name="input"
                )
                trace1 = go.Scatter(
                    x=np.arange(len(variables[:, 1])), y=out, name="output"
                )
                data = [trace0, trace1]
                plotly.offline.plot(
                    data, filename="x" + str(i + 1) + "_response.html", auto_open=True
                )

            p = np.diff(out)
            q = np.diff(trend)
            r = np.multiply(p, q)
            r_max = max(r)
            r_min = min(r)
            s = None
            if r_max <= 0 and r_min <= 0:
                s = "inverse"
            elif r_max >= 0 and r_min >= 0:
                s = "direct"
            elif r_max == 0 and r_min == 0:
                s = "nil"
            elif r_min < 0 < r_max:
                s = "mixed"

            svr = np.vstack((svr, ["x" + str(i + 1), s]))
            self.svr = svr



from math import ceil
from random import choice, random

import numpy as np
import pandas as pd
import plotly
import plotly.graph_objs as go
from graphviz import Digraph, Source

from pyrvea.EAs.PPGA import PPGA
from pyrvea.EAs.TournamentEA import TournamentEA
from pyrvea.Population.Population import Population
from pyrvea.Problem.baseproblem import BaseProblem


class BioGP(BaseProblem):
    """Creates syntax tree models to use for genetic programming through bi-objective
    genetic algorithms.

    The BioGP technique initially minimizes training error through a single objective
    optimization procedure and then a
    trade-off between complexity and accuracy is worked out through a genetic algorithm
    based bi-objective
    optimization strategy.

    The benefit of the BioGP approach is that an expert user or a decision maker (DM)
    can
    flexibly select the mathematical operations involved to construct a meta-model of
    desired complexity or
    accuracy. It is also designed to combat bloat – a perennial problem in genetic
    programming along with
    over fitting and under fitting problems.

    Notes
    -----
    The algorithm has been created earlier in MATLAB, and this Python implementation
    has been using
    that code as a basis.

    Python code has been written by Niko Rissanen under the supervision of professor
    Nirupam Chakraborti.

    Parameters
    ----------
    name : str
        Name of the problem.
    X_train : np.ndarray
        Training data input.
    y_train : np.ndarray
        Training data target values.
    num_of_objectives : int
        The number of objectives.
    params : dict
        Parameters for model training.
    num_samples : int
        The number of data points, or samples.
    function_set : array_like
        The function set to use when creating the trees.
    terminal_set : array_like
        The terminals (variables and constants) to use when creating the trees.

    References
    ----------
    [1] Giri, B. K., Hakanen, J., Miettinen, K., & Chakraborti, N. (2013). Genetic
    programming through bi-objective
    genetic algorithms with a study of a simulated moving bed process involving
    multiple objectives.
    Applied Soft Computing, 13(5), 2613-2623.
    """

    def __init__(
        self,
        name=None,
        X_train=None,
        y_train=None,
        num_of_objectives=2,
        params=None,
        num_samples=None,
        function_set=None,
        terminal_set=None,
    ):
        super().__init__()

        self.name = name
        self.X_train = X_train
        self.y_train = y_train
        self.num_of_objectives = num_of_objectives
        self.params = params
        self.num_samples = num_samples
        # These functions should be moved to their own separate package maybe
        self.function_map = {
            "add": self.add,
            "sub": self.sub,
            "mul": self.mul,
            "div": self.div,
            "sqrt": self.sqrt,
            "log": self.log,
            "sin": self.sin,
            "cos": self.cos,
            "tan": self.tan,
            "neg": self.neg,
        }
        self.terminal_set = terminal_set
        self.function_set = function_set

        self.individuals = []

    def create_individuals(self):

        if self.params["init_method"] == "ramped_half_and_half":
            for md in range(
                ceil(self.params["max_depth"] / 2), self.params["max_depth"] + 1
            ):
                for i in range(
                    int(self.params["pop_size"] / (self.params["max_depth"] + 1))
                ):

                    ind = LinearNode(value="linear", params=self.params)
                    ind.grow_tree(max_depth=md, method="grow", ind=ind)
                    self.individuals.append(ind)

                for i in range(
                    int(self.params["pop_size"] / (self.params["max_depth"] + 1))
                ):

                    ind = LinearNode(value="linear", params=self.params)
                    ind.grow_tree(max_depth=md, method="full", ind=ind)
                    self.individuals.append(ind)

        elif self.params["init_method"] == "full":
            for i in range(self.params["pop_size"]):
                ind = LinearNode(value="linear", params=self.params)
                ind.grow_tree(method="full", ind=ind)
                self.individuals.append(ind)

        elif self.params["init_method"] == "grow":
            for i in range(self.params["pop_size"]):
                ind = LinearNode(value="linear", params=self.params)
                ind.grow_tree(method="grow", ind=ind)
                self.individuals.append(ind)

        return self.individuals

    def objectives(self, decision_variables):

        return decision_variables.calculate_linear(self.X_train, self.y_train)

    def select(self, pop, non_dom_front, selection="min_error"):
        """ Select target model from the population.

        Parameters
        ----------
        pop : obj
            The population object.
        non_dom_front : list
            Indices of the models on the non-dominated front.
        selection : str
            The criterion to use for selecting the model.
            Possible values: 'min_error', 'akaike_corrected', 'manual'

        Returns
        -------
        The selected model
        """
        model = None
        fitness = None
        if selection == "min_error":
            # Return the model with the lowest error

            lowest_error = np.argmin(pop.objectives[:, 0])
            model = pop.individuals[lowest_error]
            fitness = pop.fitness[lowest_error]

        return model, fitness

    @staticmethod
    def add(x, y):
        return np.add(x, y)

    @staticmethod
    def sub(x, y):
        return np.subtract(x, y)

    @staticmethod
    def mul(x, y):
        return np.multiply(x, y)

    @staticmethod
    def div(x, y):
        y[y == 0] = 1.0
        return np.divide(x, y)

    @staticmethod
    def sqrt(x):
        return np.sqrt(np.abs(x))

    @staticmethod
    def log(x):
        return np.where(np.abs(x) > 0.001, np.log(np.abs(x)), 0.0)

    @staticmethod
    def sin(x):
        return np.sin(x)

    @staticmethod
    def cos(x):
        return np.cos(x)

    @staticmethod
    def tan(x):
        return np.tan(x)

    @staticmethod
    def neg(x):
        return np.negative(x)


class BioGPModel(BioGP):
    """The class for the BioGP surrogate model.

    Parameters
    ----------
    model_parameters : dict
        Parameters passed for the model.
    ea_parameters : dict
        Parameters passed for the genetic algorithm.

    Attributes
    ----------
    name : str
        Name of the model.
    linear_node : LinearNode
        The parent node of the tree containing the linear values.
    fitness : list
        Fitness of the trained model.
    minimize : list (bool)
        List of which objectives to minimize.
    svr : np.ndarray
        Single variable response of the model.
    log : file
        If logging set to True in params, external log file is stored here.

    """

    def __init__(self, model_parameters=None, ea_parameters=None):
        super().__init__()
        self.name = "BioGP_Model"
        self.linear_node = None
        self.fitness = None
        self.minimize = None
        self.svr = None
        self.log = None
        self.ea_params = ea_parameters
        if model_parameters:
            self.set_params(**model_parameters)
        else:
            self.set_params()

    def set_params(
        self,
        name="BioGP_Model",
        training_algorithm=PPGA,
        pop_size=500,
        max_depth=5,
        max_subtrees=4,
        prob_terminal=0.5,
        complexity_scalar=0.5,
        error_lim=0.001,
        init_method="ramped_half_and_half",
        selection="min_error",
        loss_func="root_mean_square",
        recombination_type=None,
        crossover_type="biogp_xover",
        mutation_type="biogp_mut",
        single_obj_generations=10,
        logging=False,
        plotting=False,
        function_set=("add", "sub", "mul", "div"),
        terminal_set=None,
    ):

        """ Set parameters for the BioGP model.

        Parameters
        ----------
        name : str
            Name of the problem.
        training_algorithm : EA
            Which evolutionary algorithm to use for training the models.
        pop_size : int
            Population size.
        max_depth : int
            Maximum depth of the tree.
        max_subtrees : int
            Maximum number of subtrees the tree can grow.
        prob_terminal : float
            The probability of making the node terminal when growing the tree.
        complexity_scalar : float
            Complexity of the model is calculated as a weighted aggregate of the
            maximum depth of the GP tree
            and the total number of corresponding function nodes. Larger value gives
            more weight to the depth,
            lower more weight to the number of function nodes.
        error_lim : float
            Used to control bloat. If the error reduction ratio of a subtree is less
            than this value,
            then that root is terminated and a new root is grown under the linear node
            (i.e., parent node).
        init_method : str
            Method to use for creating the initial population. Can be either 'grow',
            'full', or
            'ramped_half_and_half' (default).

            grow: nodes are chosen at random from both functions and terminals,
            allowing for smaller trees
            than max_depth.
            full: nodes are chosen from the function set until the max depth is reached,
            and then terminals are chosen.
            ramped half and half: trees are grown with a 50/50 mixture of 'grow' and
            'full'.

        loss_func : str
            The loss function to use.
        selection : str
            The selection method to use.
        recombination_type, crossover_type, mutation_type : str or None
            Recombination functions. If recombination_type is specified, crossover and
            mutation
            will be handled by the same function. If None, they are done separately.
        single_obj_generations : int
            How many generations to run minimizing only the training error.
        logging : bool
            True to create a logfile, False otherwise.
        plotting : bool
            True to create a plot, False otherwise.
        function_set : tuple
            The function set to use when creating the trees.
        terminal_set : list
            The terminals (variables and constants) to use when creating the trees.

        """

        params = {
            "name": name,
            "training_algorithm": training_algorithm,
            "pop_size": pop_size,
            "max_depth": max_depth,
            "max_subtrees": max_subtrees,
            "prob_terminal": prob_terminal,
            "complexity_scalar": complexity_scalar,
            "error_lim": error_lim,
            "init_method": init_method,
            "loss_func": loss_func,
            "selection": selection,
            "recombination_type": recombination_type,
            "crossover_type": crossover_type,
            "mutation_type": mutation_type,
            "single_obj_generations": single_obj_generations,
            "logging": logging,
            "plotting": plotting,
            "function_set": function_set,
            "terminal_set": terminal_set,
        }

        self.name = name
        self.params = params

    def fit(self, training_data, target_values):
        """Fit data in BioGP model.

        Parameters
        ----------
        training_data : pd.DataFrame, shape = (numbers of samples, number of variables)
            Training data.
        target_values : pd.DataFrame
            Target values.

        Returns
        -------
        self : returns an instance of self.

        """

        self.X_train = training_data
        self.y_train = target_values
        self.num_samples = target_values.shape[0]
        self.num_of_variables = training_data.shape[1]
        terminal_set = self.X_train.columns.tolist()
        if self.params["terminal_set"]:
            self.params["terminal_set"].extend(terminal_set)
        function_set = []
        for function in self.params["function_set"]:
            function_set.append(self.function_map[function])
        self.params["function_set"] = function_set

        self.train()

        self.single_variable_response(ploton=self.params["plotting"])

        if self.params["logging"]:
            self.create_logfile()

        return self

    def train(self):
        """Trains the networks and selects the best model from the non dominated front.

        """

        # Minimize only error for first n generations before switching to bi-objective
        ea_params = {
            "generations_per_iteration": 1,
            "iterations": self.params["single_obj_generations"],
        }
        self.minimize = [True, False]

        print(
            "Minimizing error for "
            + str(self.params["single_obj_generations"])
            + " generations..."
        )

        pop = Population(
            self,
            assign_type="BioGP",
            pop_size=self.params["pop_size"],
            plotting=self.params["plotting"],
            recombination_type=self.params["recombination_type"],
            crossover_type=self.params["crossover_type"],
            mutation_type=self.params["mutation_type"],
        )

        pop.evolve(EA=TournamentEA, ea_parameters=ea_params)

        # Switch to bi-objective (error, complexity)
        self.minimize = [True, True]
        pop.update_fitness()

        print("Switching to bi-objective mode")

        pop.evolve(EA=self.params["training_algorithm"], ea_parameters=self.ea_params)

        non_dom_front = pop.non_dominated()
        self.linear_node, self.fitness = self.select(
            pop, non_dom_front, self.params["selection"]
        )

    def predict(self, decision_variables):
        """Predict using the BioGP model.

        Parameters
        ----------
        decision_variables : pd.DataFrame
            The decision variables used for prediction.

        Returns
        -------
        y : np.ndarray
            The prediction of the model.

        """

        sub_trees = []
        # Stack outputs of subtrees to form weight matrix
        for root in self.linear_node.roots:
            sub_tree = root.predict(decision_variables)
            if sub_tree.size == 1:
                sub_tree = sub_tree.reshape(1)
            sub_trees.append(sub_tree)

        sub_trees = np.hstack(sub_trees)
        axis = None
        if sub_trees.ndim > 1:
            axis = 1
        sub_trees = np.insert(sub_trees, 0, 1, axis=axis)

        y = np.dot(sub_trees, self.linear_node.linear)

        if isinstance(y, float):
            y = np.asarray([y])

        return y

    def plot(self, prediction, target, name=None):
        """Creates and shows a plot for the model's prediction.

        Parameters
        ----------
        prediction : np.ndarray
            The prediction of the model.
        target : pd.DataFrame
            The target values.
        name : str
            Filename to save the plot as.
        """
        target = np.asarray(target)
        if name is None:
            name = self.name

        trace0 = go.Scatter(x=prediction, y=target, mode="markers")
        trace1 = go.Scatter(x=target, y=target)
        data = [trace0, trace1]
        plotly.offline.plot(
            data,
            filename="Tests/"
            + self.params["training_algorithm"].__name__
            + self.__class__.__name__
            + name
            + "_var"
            + str(self.num_of_variables)
            + "_depth"
            + str(self.params["max_depth"])
            + "_subtrees"
            + str(self.params["max_subtrees"])
            + ".html",
            auto_open=True,
        )

    def create_logfile(self, name=None):
        """Create a log file containing the parameters for training the model and the EA.

        Parameters
        ----------
        name : str
            Filename to save the log as.

        Returns
        -------
        log_file : file
            An external log file.
        """

        if name is None:
            name = self.name

        # Save params to log file
        log_file = open(
            "Tests/"
            + self.params["training_algorithm"].__name__
            + self.__class__.__name__
            + name
            + "_var"
            + str(self.num_of_variables)
            + "_depth"
            + str(self.params["max_depth"])
            + "_subtrees"
            + str(self.params["max_subtrees"])
            + ".log",
            "a",
        )

        for i in self.params:
            print("", i, ":", self.params[i], file=log_file)

        if self.fitness is not None:
            print("fitness: " + str(self.fitness), file=log_file)

        if self.svr is not None:
            print("single variable response: " + str(self.svr), file=log_file)

        return log_file

    def single_variable_response(self, ploton=False):
        """Get the model's response to a single variable.

        Parameters
        ----------
        ploton : bool
            Create and show plot on/off.
        """

        trend = np.loadtxt("trend")
        avg = np.ones((1, self.num_of_variables)) * (np.finfo(float).eps + 1) / 2
        svr = np.empty((0, 2))
        variables = np.ones((len(trend), 1)) * avg
        dataset = pd.DataFrame.from_records(variables)
        dataset.columns = self.X_train.columns

        for i in range(self.num_of_variables):

            dataset.iloc[:, i] = trend

            out = self.predict(dataset)

            if min(out) == max(out):
                out = 0.5 * np.ones(out.size)
            else:
                out = (out - min(out)) / (max(out) - min(out))

            if ploton:
                trace0 = go.Scatter(
                    x=np.arange(len(variables[:, 1])), y=variables[:, i], name="input"
                )
                trace1 = go.Scatter(
                    x=np.arange(len(variables[:, 1])), y=out, name="output"
                )
                data = [trace0, trace1]
                plotly.offline.plot(
                    data, filename="x" + str(i + 1) + "_response.html", auto_open=True
                )

            p = np.diff(out)
            q = np.diff(trend)
            r = np.multiply(p, q)
            r_max = max(r)
            r_min = min(r)
            s = None
            if r_max <= 0 and r_min <= 0:
                s = "inverse"
            elif r_max >= 0 and r_min >= 0:
                s = "direct"
            elif r_max == 0 and r_min == 0:
                s = "nil"
            elif r_min < 0 < r_max:
                s = "mixed"

            svr = np.vstack((svr, ["x" + str(i + 1), s]))
            self.svr = svr


class Node:
    """A node object representing a function or terminal node in the tree.

    Parameters
    ----------
    value : function, str or float
        A function node has as its value a function. Terminal nodes contain variables
        which are either float or str.
    depth : int
        The depth the node is at.
    params : None or dict
        The parameters of the model.
    function_set : array_like
        The function set to use when creating the trees.
    terminal_set : array_like
        The terminals (variables and constants) to use when creating the trees.

    """

    def __init__(
        self, value=None, depth=None, params=None, function_set=None, terminal_set=None
    ):
        self.value = value
        self.depth = depth
        self.params = params
        self.function_set = function_set
        self.terminal_set = terminal_set
        self.nodes = []
        self.nodes_at_depth = None
        self.total_depth = None
        self.roots = []

    def predict(self, decision_variables=None):

        if callable(self.value):
            values = [root.predict(decision_variables) for root in self.roots]
            if self.depth == 0:
                self.value = sum(values)
                return self.value
            else:
                return self.value(*values)

        else:
            if isinstance(decision_variables, np.ndarray) and isinstance(
                self.value, str
            ):
                return decision_variables[
                    :, int("".join(filter(str.isdigit, self.value))) - 1
                ]
            if isinstance(self.value, str):
                return np.asarray(decision_variables[self.value]).reshape(-1, 1)
            else:
                return np.full((decision_variables.shape[0], 1), self.value)

    def node_label(self):  # return string label
        if callable(self.value):
            return self.value.__name__
        else:
            return str(self.value)

    def draw(self, dot, count):  # dot & count are lists in order to pass "by reference"
        node_name = str(count[0])
        dot[0].node(node_name, self.node_label())

        for node in self.roots:
            count[0] += 1
            dot[0].edge(node_name, str(count[0]))
            node.draw(dot, count)

    def draw_tree(self, name="tree", footer=""):
        dot = [Digraph()]
        dot[0].attr(kw="graph", label=footer)
        count = [0]
        self.draw(dot, count)
        Source(dot[0], filename=name + ".gv", format="png").render()

    def get_sub_nodes(self):
        """Get all nodes belonging to the subtree under the current node.

        Returns
        -------
        nodes : list
            A list of nodes in the subtree.

        """
        nodes = []
        nodes_at_depth = {}
        stack = [self]
        while stack:
            cur_node = stack[0]
            depth = cur_node.depth
            if cur_node.depth not in nodes_at_depth:
                nodes_at_depth[cur_node.depth] = []
            nodes_at_depth[cur_node.depth].append(cur_node)
            stack = stack[1:]
            nodes.append(cur_node)
            if len(cur_node.roots) > 0:
                depth += 1
            for child in cur_node.roots:
                child.depth = depth
                stack.append(child)

        self.nodes_at_depth = nodes_at_depth
        self.total_depth = max(key for key in self.nodes_at_depth.keys())

        return nodes

    def grow_tree(self, max_depth=None, method="grow", depth=0, ind=None):
        """Create a random tree recursively using either grow or full method.

        Parameters
        ----------
        max_depth : int
            The maximum depth of the tree.
        method : str
            Methods: 'grow', 'full'.
            For the 'grow' method, nodes are chosen at random from both functions and
            terminals.
            The 'full' method chooses nodes from the function set until the max depth
            is reached,
            and then terminals are chosen.
        depth : int
            Current depth.
        ind : :obj:
            The starting node from which to begin growing trees.

        """
        node = None
        if max_depth is None:
            max_depth = self.params["max_depth"]
        if depth == 0:
            if ind is None:
                ind = LinearNode(value="linear")
            num_subtrees = self.params["max_subtrees"]
            for i in range(len(ind.roots), num_subtrees):
                node = self.grow_tree(max_depth, method, depth=depth + 1)
                ind.roots.append(node)

        # Make terminal node
        elif (
            depth >= max_depth
            or method == "grow"
            and random() < self.params["prob_terminal"]
        ):
            node = Node(
                depth=depth,
                function_set=self.params["function_set"],
                terminal_set=self.params["terminal_set"],
            )
            node.value = choice(node.terminal_set)

        # Make function node
        else:
            node = Node(
                depth=depth,
                function_set=self.params["function_set"],
                terminal_set=self.params["terminal_set"],
            )
            node.value = choice(node.function_set)

            for i in range(node.value.__code__.co_argcount):  # Check arity
                root = self.grow_tree(max_depth, method, depth=depth + 1)
                node.roots.append(root)

        return node


class LinearNode(Node):
    """The parent node of the tree, from which a number of subtrees emerge, as defined
    by the user. The linear node takes a weighted sum of the output from the subtrees
    and
    also uses a bias value. The weights and the bias are calculated by the linear least
    square technique.

    Parameters
    ----------
    value : function, str or float
        A function node has as its value a function. Terminal nodes contain variables
        which are either float or str.
    depth : int
        The depth the node is at.
    params : None or dict
        Parameters of the model.
    """

    def __init__(self, value="linear", depth=0, params=None):
        super().__init__(params=params)
        self.value = value
        self.nodes_at_depth = {}
        self.params = params
        self.depth = depth
        self.total_depth = None
        self.out = None
        self.complexity = None
        self.fitness = None
        self.linear = None

    def calculate_linear(self, X_train, y_train):

        sub_trees = []
        # Stack outputs of subtrees to form weight matrix
        for root in self.roots:
            sub_trees.append(root.predict(X_train))

        sub_trees = np.hstack(sub_trees)
        axis = None
        if sub_trees.ndim > 1:
            axis = 1
        sub_trees = np.insert(sub_trees, 0, 1, axis=axis)

        weights, *_ = np.linalg.lstsq(sub_trees, y_train, rcond=None)
        self.linear = weights
        out = np.dot(sub_trees, weights)

        # Error reduction ratio
        q, r = np.linalg.qr(sub_trees)
        s = np.linalg.lstsq(q, y_train, rcond=None)[0]
        error = np.divide((s ** 2 * np.sum(q * q, axis=0)), np.sum(out * out, axis=0))

        # If error reduction ration < err_lim, delete root and grow new one
        for i, err in enumerate(error[1:]):
            if err < self.params["error_lim"]:
                del self.roots[i]
                self.grow_tree(
                    max_depth=self.params["max_depth"], method="grow", ind=self
                )

        self.nodes = self.get_sub_nodes()

        num_func_nodes = sum(1 for node in self.nodes if callable(node.value))

        complexity = (
            self.params["complexity_scalar"] * self.total_depth
            + (1 - self.params["complexity_scalar"]) * num_func_nodes
        )

        self.out = out
        self.complexity = complexity
        training_error = None

        if self.params["loss_func"] == "root_mean_square":
            training_error = np.sqrt(np.mean(((y_train - out) ** 2)))

        elif self.params["loss_func"] == "root_median_square":
            training_error = np.sqrt(np.median(((y_train - out) ** 2)))

        return [training_error, self.complexity]



from typing import List

import numpy as np
import pandas as pd
from sklearn import preprocessing
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.metrics import r2_score
from sklearn.model_selection import train_test_split as tts
from sklearn.neural_network import MLPRegressor

from pyrvea.Problem.baseproblem import BaseProblem
from pyrvea.Problem.biogp_problem import BioGPModel as BioGP
from pyrvea.Problem.evodn2_problem import EvoDN2Model as EvoDN2
from pyrvea.Problem.evonn_problem import EvoNNModel as EvoNN


class DataProblem(BaseProblem):
    def __init__(
        self,
        data: pd.DataFrame = None,
        x: List[str] = None,
        y: List[str] = None,
        minimize: List[bool] = None,
        ideal: List[float] = None,
        nadir: List[float] = None,
        num_of_constraints=0,
        lower_limits: List[float] = None,
        upper_limits: List[float] = None,
        name="data_problem",
    ):
        super().__init__()
        self.raw_data = data
        self.data = data
        self.x = x
        self.y = y
        self.num_of_variables = len(x)
        self.num_of_objectives = len(y)
        self.num_of_constraints = num_of_constraints
        self.number_of_samples = self.data.shape[0]
        self.name = name
        if minimize is None:
            self.minimize = [True] * self.num_of_objectives
        else:
            assert len(minimize) == self.num_of_objectives
            self.minimize = minimize
        self.preprocessing_transformations = []
        # These indices define how the training will occur.
        # all_indices is the list of indices that can be used. This does not include
        # outliers.
        self.all_indices = list(range(self.number_of_samples))
        # These should be list of lists.
        # The inner list contains the list of indices.
        # The outer list, if it contains multiple elements, should lead to multiple
        # trained models. Can be used for k-fold validation.
        # Train and test indices super list should have the same length.
        self.train_indices: List[List] = []
        self.test_indices: List[List] = []
        self.validation_indices: List[List] = None

        # self.models = dict.fromkeys(self.y, [])  # This method duplicates models in each key, why?
        self.models = {}                           # This works..
        for obj in self.y:
            self.models[obj] = []

        self.metrics = []
        # Defining bounds in the decision space
        if lower_limits is None:
            self.lower_limits = np.min(self.raw_data[x], axis=0)
        else:
            assert len(lower_limits) == self.num_of_variables
            self.lower_limits = lower_limits
        if upper_limits is None:
            self.upper_limits = np.max(self.raw_data[x], axis=0)
        else:
            assert len(upper_limits) == self.num_of_variables
            self.upper_limits = upper_limits

    def data_scaling(self, data_decision):  # Scales the data from 0 to 1
        # Check this range stuff
        min_max_scaler = preprocessing.MinMaxScaler(
            feature_range=(self.lower_limits, self.upper_limits)
        )
        self.processed_data_decision = min_max_scaler.fit_transform(data_decision)
        self.preprocessing_transformations.append(min_max_scaler)

    def data_uniform_mapping(self):  # Maps the data to uniform distribution
        pass

    def outlier_removal(self):  # Removes the outliers
        pass

    def train_test_split(self, train_size: float = 0.8):  # Split dataset

        for x in range(1):
            train_indices, test_indices = tts(self.all_indices, train_size=train_size)
            train_indices.sort()
            test_indices.sort()
            self.train_indices.append(train_indices)
            self.test_indices.append(test_indices)

    def train(self, model_type: str = None, objectives: str = None, **kwargs):
        if objectives is None:
            objectives = self.y
        if model_type is None:
            model_type = "MLP"
        surrogate_model_options = {
            "GPR": GaussianProcessRegressor,
            "MLP": MLPRegressor,
            "EvoNN": EvoNN,
            "EvoDN2": EvoDN2,
            "BioGP": BioGP
        }
        model_type = surrogate_model_options[model_type]
        # Build specific surrogate models
        print("Building Surrogate Models ...")
        # Fit to data using Maximum Likelihood Estimation of the parameters
        for obj in objectives:

            print("Building model for " + str(obj))
            for train_run, train_indices in enumerate(self.train_indices):
                print("Training run number", train_run, "of", len(self.train_indices))
                model = model_type(**kwargs)
                model.fit(
                    self.data[self.x].loc[train_indices],
                    self.data[obj].loc[train_indices],
                )
                self.models[obj].append(model)

        # Select model
        print("Surrogate models build completed.")

    def transform_new_data(self, decision_variables):
        decision_variables_transformed = decision_variables
        if len(self.preprocessing_transformations) > 0:
            for transformation in self.preprocessing_transformations:
                decision_variables_transformed = transformation.transfrom(
                    decision_variables_transformed
                )
        return decision_variables_transformed

    def surrogates_predict(self, decision_variables):
        y_pred = None
        # transforms applied
        decision_variables_transformed = self.transform_new_data(decision_variables)
        for obj, i in zip(self.y, range(self.num_of_objectives)):
            y = self.models[obj][0].predict(
                decision_variables_transformed
            ).reshape(-1, 1)
            if y_pred is None:
                y_pred = np.asarray(y)
            else:
                y_pred = np.hstack((y_pred, y))
        return y_pred

    def testing_score(self):  # Return R-squared of testing
        x, y = self.select_data(self.test_indices)
        x = self.transform_new_data(x)
        y_pred = None
        for i in range(np.shape(x)[0]):
            if y_pred is None:
                y_pred = np.asarray(self.surrogates_predict(x[i]))
            else:
                y_pred = np.vstack((y_pred, self.surrogates_predict(x[i])))
        for i in range(np.shape(y)[1]):
            self.r_sq.append(r2_score(y[:, i], y_pred[:, i]))
        return y_pred

    def retrain_surrogate(self):
        pass

    def objectives(self, decision_variables):
        """Objectives function to use in optimization.

        Parameters
        ----------
        decision_variables : ndarray
            The decision variables

        Returns
        -------
        objectives : ndarray
            The objective values

        """
        objectives = []
        for obj in self.y:
            objectives.append(
                self.models[obj][0].predict(decision_variables.reshape(1, -1))[0]
            )

        return objectives



import random

import numpy as np
import plotly
import plotly.graph_objs as go
from scipy.special import expit

from pyrvea.EAs.PPGA import PPGA
from pyrvea.Population.Population import Population
from pyrvea.Problem.baseproblem import BaseProblem


class EvoDN2(BaseProblem):
    """Creates Deep Neural Networks (DNN) for the EvoDN2 algorithm.

    DNNs have a fixed number of subnets, each of which has a random number of
    layers, and a random number of nodes in each layer, dependant on max layers
    and max nodes set by user.

    Notes
    -----
    The algorithm has been created earlier in MATLAB, and this Python implementation
    has been using
    that code as a basis.

    Python code has been written by Niko Rissanen under the supervision of professor
    Nirupam Chakraborti.

    Parameters
    ----------
    name : str
        Name of the problem
    X_train : ndarray
        Training data input
    y_train : ndarray
        Training data target values
    num_of_objectives : int
        The number of objectives
    num_samples : int
        The number of data points, or samples.
    subsets : array_like
        A list of variables used for each subnet of the model.
    params : dict
        Parameters for the models

    References
    ----------
    [1] Swagata R., Bhupinder S., Chakrabarti, N. and Chakraborti, N. (2019). A new
    Deep Neural Network algorithm
    employed in the study of mechanical properties of micro-alloyed steel.
    Department of Metallurgical and Materials Engineering, Indian Institute of
    Technology.
    """

    def __init__(
        self,
        name=None,
        X_train=None,
        y_train=None,
        num_of_objectives=2,
        num_samples=None,
        subsets=None,
        params=None,
    ):
        super().__init__()

        self.name = name
        self.X_train = X_train
        self.y_train = y_train
        self.num_of_objectives = num_of_objectives
        self.num_samples = num_samples
        self.subsets = subsets
        self.params = params

    def objectives(self, decision_variables) -> list:
        """ Use this method to calculate objective functions.

        Parameters
        ----------
        decision_variables : ndarray
            Variables from the neural network
        Returns
        -------
        obj_func : list
            The objective function
        """

        non_linear_layer, complexity = self.activation(decision_variables)
        _, predicted_values, training_error = self.calculate_linear(non_linear_layer)

        obj_func = [training_error, complexity]

        return obj_func

    def activation(self, decision_variables):
        """ Calculates the dot product and applies the activation function.
        Parameters
        ----------
        decision_variables : ndarray
            Array of all subnets in the current neural network

        Returns
        -------
        non_linear_layer : ndarray
            The final non-linear layer before the output
        complexity : float
            The complexity of the neural network
        """
        network_complexity = []
        non_linear_layer = np.empty((self.num_samples, 0))

        for i, subnet in enumerate(decision_variables):

            # Get the input variables for the first layer
            in_nodes = self.X_train[:, self.subsets[i]]
            subnet_complexity = 1

            for layer in subnet:
                # Calculate the dot product + bias
                out = np.dot(in_nodes, layer[1:, :]) + layer[0]

                subnet_complexity = np.dot(subnet_complexity, np.abs(layer[1:, :]))

                in_nodes = self.activate(self.params["activation_func"], out)

            network_complexity.append(np.sum(subnet_complexity))
            non_linear_layer = np.hstack((non_linear_layer, in_nodes))

        complexity = np.sum(network_complexity)

        return non_linear_layer, complexity

    def calculate_linear(self, activated_layer):
        """ Apply the linear function to the final output layer
        and calculate the training error.

        Parameters
        ----------
        activated_layer : ndarray
            Output of the activation function.

        Returns
        -------
        linear_layer : ndarray
            The optimized output layer of the network.
        predicted_values : ndarray
            The prediction of the model.
        training_error : float
            Training error of the model.
        """

        linear_layer = None
        predicted_values = None
        training_error = None

        if self.params["opt_func"] == "llsq":
            linear_solution = np.linalg.lstsq(activated_layer, self.y_train, rcond=None)
            linear_layer = linear_solution[0]
            predicted_values = np.dot(activated_layer, linear_layer)

        if self.params["loss_func"] == "root_mean_square":
            training_error = np.sqrt(np.mean(((self.y_train - predicted_values) ** 2)))

        elif self.params["loss_func"] == "root_median_square":
            training_error = np.sqrt(
                np.median(((self.y_train - predicted_values) ** 2))
            )

        return linear_layer, predicted_values, training_error

    def select(self, pop, non_dom_front, selection="min_error"):
        """ Select target model from the population.

        Parameters
        ----------
        pop : obj
            The population object.
        non_dom_front : list
            Indices of the models on the non-dominated front.
        selection : str
            The criterion to use for selecting the model.
            Possible values: 'min_error', 'manual'

        Returns
        -------
        model:
            The selected model.
        fitness:
            The model's fitness values.

        """
        model = None
        fitness = None
        if selection == "min_error":
            # Return the model with the lowest error

            lowest_error = np.argmin(pop.objectives[:, 0])
            model = pop.individuals[lowest_error]
            fitness = pop.fitness[lowest_error]

        elif selection == "manual":

            pareto = pop.objectives[non_dom_front]
            hover = pop.objectives[non_dom_front].tolist()

            for i, x in enumerate(hover):
                x.insert(0, "Model " + str(non_dom_front[i]) + "<br>")

            trace0 = go.Scatter(
                x=pop.objectives[:, 0], y=pop.objectives[:, 1], mode="markers"
            )
            trace1 = go.Scatter(
                x=pareto[:, 0],
                y=pareto[:, 1],
                text=hover,
                hoverinfo="text",
                mode="markers+lines",
            )
            data = [trace0, trace1]
            layout = go.Layout(
                xaxis=dict(title="training error"), yaxis=dict(title="complexity")
            )
            plotly.offline.plot(
                {"data": data, "layout": layout},
                filename=self.name + "_training_models_" + "pareto" + ".html",
                auto_open=True,
            )

            model_idx = None
            while model_idx not in non_dom_front:
                usr_input = input(
                    "Please input the number of the model of your preference: "
                )
                try:
                    model_idx = int(usr_input)
                except ValueError:
                    print("Invalid input, please enter the model number.")
                    continue

                if model_idx not in non_dom_front:
                    print("Model " + str(model_idx) + " not found.")

            model = pop.individuals[int(model_idx)]
            fitness = pop.fitness[int(model_idx)]

        return model, fitness

    @staticmethod
    def activate(name, x):
        if name == "sigmoid":
            return expit(x)

        if name == "relu":
            return np.maximum(x, 0)

        if name == "tanh":
            return np.tanh(x)


class EvoDN2Model(EvoDN2):
    """Class for the EvoDN2 surrogate model.

    Parameters
    ----------
    model_parameters : dict
        Parameters passed for the model.
    ea_parameters : dict
        Parameters passed for the genetic algorithm.

    Attributes
    ----------
    name : str
        Name of the problem
    subnets : array_like
        A list containing the subnets of the model.
    subsets : array_like
        A list of variables used for each subnet of the model.
    fitness : list
        Fitness of the trained model.
    non_linear_layer : np.ndarray or None
        The activated layer combining all of the model's subnets.
    linear_layer : np.ndarray or None
        The final optimized layer of the network.
    svr : array_like
        Single variable response of the model.
    log : file
        If logging set to True in params, external log file is stored here.

    """

    def __init__(self, model_parameters=None, ea_parameters=None):
        super().__init__()
        self.name = "EvoDN2_Model"
        self.subnets = None
        self.subsets = None
        self.fitness = []
        self.non_linear_layer = None
        self.linear_layer = None
        self.svr = None
        self.log = None
        self.ea_params = ea_parameters
        if model_parameters:
            self.set_params(**model_parameters)
        else:
            self.set_params()

    def set_params(
        self,
        name="EvoDN2_Model",
        training_algorithm=PPGA,
        pop_size=500,
        num_subnets=4,
        max_layers=4,
        max_nodes=4,
        prob_omit=0.2,
        w_low=-5.0,
        w_high=5.0,
        activation_func="sigmoid",
        opt_func="llsq",
        loss_func="root_mean_square",
        selection="min_error",
        recombination_type="evodn2_xover_mutation",
        crossover_type="standard",
        mutation_type="gaussian",
        logging=False,
        plotting=False,
    ):
        """ Set parameters for EvoDN2 model.

        Parameters
        ----------
        name : str
            Name of the problem.
        training_algorithm : EA
            Which training algorithm to use.
        pop_size : int
            Population size.
        num_subnets : int
            Number of subnets.
        max_layers : int
            Maximum number of hidden layers in each subnet.
        max_nodes : int
            Maximum number of nodes in each hidden layer.
        prob_omit : float
            Probability of setting some weights to zero initially.
        w_low : float
            The lower bound for randomly generated weights.
        w_high : float
            The upper bound for randomly generated weights.
        activation_func : str
            Function to use for activation.
        opt_func : str
            Function to use for optimizing the final layer of the model.
        loss_func : str
            The loss function to use.
        selection : str
            The selection to use for selecting the model.
        recombination_type, crossover_type, mutation_type : str
            Recombination functions. If recombination_type is specified, crossover and
            mutation
            will be handled by the same function. If None, they are done separately.
        logging : bool
            True to create a logfile, False otherwise.
        plotting : bool
            True to create a plot, False otherwise.

        """
        params = {
            "name": name,
            "training_algorithm": training_algorithm,
            "pop_size": pop_size,
            "num_subnets": num_subnets,
            "max_layers": max_layers,
            "max_nodes": max_nodes,
            "prob_omit": prob_omit,
            "w_low": w_low,
            "w_high": w_high,
            "activation_func": activation_func,
            "opt_func": opt_func,
            "loss_func": loss_func,
            "selection": selection,
            "crossover_type": crossover_type,
            "mutation_type": mutation_type,
            "recombination_type": recombination_type,
            "logging": logging,
            "plotting": plotting,
        }

        self.name = name
        self.params = params

    def fit(self, training_data, target_values):
        """Fit data in EvoDN2 model, divide input variables for each subnet randomly,
        and train the model.

        Parameters
        ----------
        training_data : pd.DataFrame, shape = (numbers of samples, number of variables)
            Training data.
        target_values : pd.DataFrame
            Target values.

        Returns
        -------
        self : returns an instance of self.

        """
        self.X_train = np.asarray(training_data)
        self.y_train = np.asarray(target_values)
        self.num_samples = target_values.shape[0]
        self.num_of_variables = training_data.shape[1]

        self.subsets = []

        # Create random subsets of decision variables for each subnet

        for i in range(self.params["num_subnets"]):
            n = random.randint(1, self.X_train.shape[1])
            self.subsets.append(random.sample(range(self.X_train.shape[1]), n))

        # Ensure that each decision variable is used as an input in at least one subnet
        for n in list(range(self.X_train.shape[1])):
            if not any(n in k for k in self.subsets):
                self.subsets[random.randint(0, self.params["num_subnets"] - 1)].append(
                    n
                )

        self.train()

        self.single_variable_response(ploton=self.params["plotting"])

        if self.params["logging"]:
            self.create_logfile()

        return self

    def train(self):
        """
        Create a random population, evolve it and select a model based on selection.
        """
        pop = Population(
            self,
            assign_type="EvoDN2",
            pop_size=self.params["pop_size"],
            recombination_type=self.params["recombination_type"],
            crossover_type=self.params["crossover_type"],
            mutation_type=self.params["mutation_type"],
            plotting=self.params["plotting"],
        )

        pop.evolve(EA=self.params["training_algorithm"], ea_parameters=self.ea_params)

        non_dom_front = pop.non_dominated()
        self.subnets, self.fitness = self.select(
            pop, non_dom_front, self.params["selection"]
        )
        self.non_linear_layer, _ = self.activation(self.subnets)
        self.linear_layer, *_ = self.calculate_linear(self.non_linear_layer)

    def predict(self, decision_variables):
        """Predict using the EvoDN2 model.

        Parameters
        ----------
        decision_variables : pd.DataFrame
            The decision variables used for prediction.

        Returns
        -------
        y : ndarray
            The prediction of the model.

        """
        non_linear_layer = np.empty((decision_variables.shape[0], 0))

        for i, subnet in enumerate(self.subnets):

            in_nodes = np.asarray(decision_variables)[:, self.subsets[i]]

            for layer in subnet:

                out = np.dot(in_nodes, layer[1:, :]) + layer[0]

                in_nodes = self.activate(self.params["activation_func"], out)

            non_linear_layer = np.hstack((non_linear_layer, in_nodes))

        y = np.dot(non_linear_layer, self.linear_layer)

        return y

    def plot(self, prediction, target, name=None):
        """Creates and shows a plot for the model's prediction.

        Parameters
        ----------
        prediction : np.ndarray
            The prediction of the model.
        target : pd.DataFrame
            The target values.
        name : str
            Filename to save the plot as.
        """
        target = np.asarray(target)
        if name is None:
            name = self.name

        trace0 = go.Scatter(x=prediction, y=target, mode="markers")
        trace1 = go.Scatter(x=target, y=target)
        data = [trace0, trace1]
        plotly.offline.plot(
            data,
            filename="Tests/"
            + self.params["training_algorithm"].__name__
            + self.__class__.__name__
            + name
            + "_var"
            + str(self.num_of_variables)
            + "_nodes"
            + str(self.params["num_subnets"])
            + "_"
            + str(self.params["max_layers"])
            + "_"
            + str(self.params["max_nodes"])
            + ".html",
            auto_open=True,
        )

    def create_logfile(self, name=None):
        """Create a log file containing the parameters for training the model and the EA.

        Returns
        -------
        log_file : file
            An external log file.
        """

        if name is None:
            name = self.name

        # Save params to log file
        log_file = open(
            "Tests/"
            + self.params["training_algorithm"].__name__
            + self.__class__.__name__
            + name
            + "_var"
            + str(self.num_of_variables)
            + "_nodes"
            + str(self.params["num_subnets"])
            + "_"
            + str(self.params["max_layers"])
            + "_"
            + str(self.params["max_nodes"])
            + ".log",
            "a",
        )

        for i in self.params:
            print("", i, ":", self.params[i], file=log_file)

        if self.fitness is not None:
            print("fitness: " + str(self.fitness), file=log_file)

        if self.svr is not None:
            print("single variable response: " + str(self.svr), file=log_file)

        return log_file

    def single_variable_response(self, ploton=False):
        """Get the model's response to a single variable.

        Parameters
        ----------
        ploton : bool
            Create and show plot on/off.

        """

        trend = np.loadtxt("trend")
        avg = np.ones((1, self.num_of_variables)) * (np.finfo(float).eps + 1) / 2
        svr = np.empty((0, 2))
        variables = np.ones((len(trend), 1)) * avg

        for i in range(self.num_of_variables):

            variables[:, i] = trend

            out = self.predict(variables)

            if min(out) == max(out):
                out = 0.5 * np.ones(out.size)
            else:
                out = (out - min(out)) / (max(out) - min(out))

            if ploton:
                trace0 = go.Scatter(
                    x=np.arange(len(variables[:, 1])), y=variables[:, i], name="input"
                )
                trace1 = go.Scatter(
                    x=np.arange(len(variables[:, 1])), y=out, name="output"
                )
                data = [trace0, trace1]
                plotly.offline.plot(
                    data, filename="x" + str(i + 1) + "_response.html", auto_open=True
                )

            p = np.diff(out)
            q = np.diff(trend)
            r = np.multiply(p, q)
            r_max = max(r)
            r_min = min(r)
            s = None
            if r_max <= 0 and r_min <= 0:
                s = "inverse"
            elif r_max >= 0 and r_min >= 0:
                s = "direct"
            elif r_max == 0 and r_min == 0:
                s = "nil"
            elif r_min < 0 < r_max:
                s = "mixed"

            svr = np.vstack((svr, ["x" + str(i + 1), s]))
            self.svr = svr



import numpy as np
import pandas as pd
from optproblems import dtlz, zdt
from pyDOE import lhs
from sklearn.preprocessing import minmax_scale

from pyrvea.Problem.baseproblem import BaseProblem
from pyrvea.Problem.test_functions import OptTestFunctions


class TestProblem(BaseProblem):
    """Test functions for single/multi-objective problems to test
    the performance of evolutionary algorithms.

    See: https://en.wikipedia.org/wiki/Test_functions_for_optimization

    Parameters
    ----------
    name : str
        Name of the test function.
    num_of_variables : int
        Number of decision variables.
    num_of_objectives : int
        Number of objectives.
    num_of_constraints : int
        Number of constraints.
    upper_limits : float
        Upper boundaries for test data.
    lower_limits : float
        Lower boundaries for test data.
    """

    def __init__(
        self,
        name=None,
        num_of_variables=None,
        num_of_objectives=None,
        num_of_constraints=0,
        upper_limits=1.0,
        lower_limits=0.0,
    ):

        super().__init__(
            name,
            num_of_variables,
            num_of_objectives,
            num_of_constraints,
            upper_limits,
            lower_limits,
        )
        if name == "ZDT1":
            self.obj_func = zdt.ZDT1()

        elif name == "ZDT2":
            self.obj_func = zdt.ZDT2()

        elif name == "ZDT3":
            self.obj_func = zdt.ZDT3()

        elif name == "ZDT4":
            self.obj_func = zdt.ZDT4()
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        elif name == "ZDT5":
            self.obj_func = zdt.ZDT5()

        elif name == "ZDT6":
            self.obj_func = zdt.ZDT6()

        elif name == "DTLZ1":
            self.obj_func = dtlz.DTLZ1(num_of_objectives, num_of_variables)
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        elif name == "DTLZ2":
            self.obj_func = dtlz.DTLZ2(num_of_objectives, num_of_variables)
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        elif name == "DTLZ3":
            self.obj_func = dtlz.DTLZ3(num_of_objectives, num_of_variables)
            self.lower_limits = 0
            self.upper_limits = 1
        elif name == "DTLZ4":
            self.obj_func = dtlz.DTLZ4(num_of_objectives, num_of_variables)
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        elif name == "DTLZ5":
            self.obj_func = dtlz.DTLZ5(num_of_objectives, num_of_variables)
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        elif name == "DTLZ6":
            self.obj_func = dtlz.DTLZ6(num_of_objectives, num_of_variables)
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        elif name == "DTLZ7":
            self.obj_func = dtlz.DTLZ7(num_of_objectives, num_of_variables)
            self.lower_limits = self.obj_func.min_bounds
            self.upper_limits = self.obj_func.max_bounds
        else:
            self.obj_func = OptTestFunctions(
                name=self.name, num_of_variables=num_of_variables
            )
            self.lower_limits = self.obj_func.lower_limits
            self.upper_limits = self.obj_func.upper_limits
            self.num_of_variables = self.obj_func.num_of_variables
            self.num_of_objectives = self.obj_func.num_of_objectives

    def objectives(self, decision_variables) -> list:
        """Use this method to calculate objective functions.

        Args:
            decision_variables:
        """
        return self.obj_func(decision_variables)

    def constraints(self, decision_variables, objective_variables):
        """Calculate constraint violation.

        Args:
            decision_variables:
            objective_variables:
        """
        print("Error: Constraints not supported yet.")

    def create_training_data(self, samples=500, method="random", seed=None):
        """Create training data for test functions.

        Parameters
        ----------
        samples : int
            Number of samples.
        method : str
            Method to use in data creation. Possible values random, lhs (latin
            hypercube sampling), linear.
        seed : int
            If a number is given, random data will be seeded.

        Returns
        -------
        dataset : Training data in a Pandas DataFrame, including all variables and
        objectives.
        x : List of variables
        y : List of objectives

        """

        np.random.seed(seed)
        training_data_input = None

        if method == "random":

            training_data_input = np.random.uniform(
                self.lower_limits, self.upper_limits, (samples, self.num_of_variables)
            )

        elif method == "lhs":
            # Latin Hypercube Sampling

            training_data_input = lhs(self.num_of_variables, samples)
            training_data_input = np.round(
                minmax_scale(
                    training_data_input, (self.lower_limits, self.upper_limits)
                ),
                decimals=5,
            )

        elif method == "linear":

            training_data_input = np.linspace(
                self.lower_limits, self.upper_limits, samples
            )

        training_data_output = np.asarray(
            [self.objectives(x) for x in training_data_input]
        )
        if self.num_of_objectives == 1:
            training_data_output = training_data_output[:, None]

        # Convert numpy array into pandas dataframe, and make columns for it
        data = np.hstack((training_data_input, training_data_output))
        dataset = pd.DataFrame.from_records(data)
        x = []
        y = []
        for var in range(training_data_input.shape[1]):
            x.append("x" + str(var + 1))
        for obj in range(training_data_output.shape[1]):
            y.append("f" + str(obj + 1))
        dataset.columns = x + y

        np.random.seed(None)

        return dataset, x, y



from typing import Dict

from desdeo_emo.EAs.BaseEA import BaseDecompositionEA
from desdeo_emo.population.Population import Population
from desdeo_emo.selection.NSGAIII_select import NSGAIII_select
from desdeo_problem.Problem import MOProblem


class NSGAIII(BaseDecompositionEA):
    """Python Implementation of NSGA-III. Based on the pymoo package.

    Most of the relevant code is contained in the super class. This class just assigns
    the NSGAIII selection operator to BaseDecompositionEA.
    
    Parameters
    ----------
    problem : MOProblem
        The problem class object specifying the details of the problem.
    population_size : int, optional
        The desired population size, by default None, which sets up a default value
        of population size depending upon the dimensionaly of the problem.
    population_params : Dict, optional
        The parameters for the population class, by default None. See
        desdeo_emo.population.Population for more details.
    initial_population : Population, optional
        An initial population class, by default None. Use this if you want to set up
        a specific starting population, such as when the output of one EA is to be
        used as the input of another.
    lattice_resolution : int, optional
        The number of divisions along individual axes in the objective space to be
        used while creating the reference vector lattice by the simplex lattice
        design. By default None
    a_priori : bool, optional
        A bool variable defining whether a priori preference is to be used or not.
        By default False
    interact : bool, optional
        A bool variable defining whether interactive preference is to be used or
        not. By default False
    n_iterations : int, optional
        The total number of iterations to be run, by default 10. This is not a hard
        limit and is only used for an internal counter.
    n_gen_per_iter : int, optional
        The total number of generations in an iteration to be run, by default 100.
        This is not a hard limit and is only used for an internal counter.
    total_function_evaluations :int, optional
        Set an upper limit to the total number of function evaluations. When set to
        zero, this argument is ignored and other termination criteria are used.
    """

    def __init__(
        self,
        problem: MOProblem,
        population_size: int = None,
        population_params: Dict = None,
        n_survive: int = None,
        initial_population: Population = None,
        lattice_resolution: int = None,
        selection_type: str = None,
        a_priori: bool = False,
        interact: bool = False,
        use_surrogates: bool = False,
        n_iterations: int = 10,
        n_gen_per_iter: int = 100,
        total_function_evaluations: int = 0,
    ):
        super().__init__(
            problem=problem,
            population_size=population_size,
            population_params=population_params,
            initial_population=initial_population,
            lattice_resolution=lattice_resolution,
            a_priori=a_priori,
            interact=interact,
            use_surrogates=use_surrogates,
            n_iterations=n_iterations,
            n_gen_per_iter=n_gen_per_iter,
            total_function_evaluations=total_function_evaluations,
        )
        self.selection_type = selection_type
        selection_operator = NSGAIII_select(
            self.population, n_survive, selection_type=selection_type
        )
        self.selection_operator = selection_operator
